/*
 * palette.c -- Functions relating to color palettes.
 */

#include "muse.h"

#define LARGE_NUMBER 2000000L;

#if XVT_CC_PROTO
MUSE_API palette_construct(FILE *file, PALETTE **pointer)
#else
MUSE_API palette_construct(file, pointer)
    FILE           *file;
    PALETTE       **pointer;
#endif
{
    PALETTE        *palette;
    ERRSTATUS       status;
    size_t          items_read = 0;
    unsigned char   palrec[SZ_PALETTE];
    long            temp = SZ_PALETTE;
    char            string[80];

    palette = (PALETTE *) muse_malloc(sizeof(PALETTE), "palette");
    if (palette == (PALETTE *) NULL)
	return (STAT_MEM_ALLOC_ERR);
    *pointer = palette;
    memset(palette, '\0', sizeof(PALETTE));
    palette->magic = MAGIC_PALETTE;

#if XVT_OS == XVT_OS_WIN
    if (temp < sizeof(PALETTE))
	printf("Reset SZ_PALETTE to:%u", sizeof(PALETTE));
#endif

    if (file)
    {
	sprintf(string, "reading pal: file pos:%ld", ftell(file));
	/*muse_log(string);*/
	items_read = fread((void *) palrec, sizeof(palrec), (size_t) 1, file);
	decode_palette(palrec, palette);
	if (items_read != 1)
	{
	    printf("Unable to read in the palette");
	    return (STAT_FILE_READ_ERR);
	}
	if (palette->magic != MAGIC_PALETTE)
	{
	    printf("The palette data read in is incorrect.");
	    return (STAT_DATA_ERR);
	}
	*pointer = palette;
	return STAT_SUCCESS;
    }
    else
    {
	palette->type = PAL_NUM_BINS;
	palette->bins.r = 6;
	palette->bins.g = 6;
	palette->bins.b = 6;

	status = pal_gen(palette);
	if (status)
	    printf("Xvt_Error in pal_gen");
    }

    return STAT_SUCCESS;
}				/* end palette_construct */


#if XVT_CC_PROTO
MUSE_API palette_destruct(
		 FILE * file,
		 BOOLEAN destruct,
		 PALETTE ** palette)
#else
MUSE_API palette_destruct(file, destruct, palette)
    FILE           *file;
    BOOLEAN         destruct;
    PALETTE       **palette;
#endif
{
    size_t          items_written = 0;
    unsigned char   palrec[SZ_PALETTE];
    char            string[80];

    if (*palette == (PALETTE *) NULL)
	return STAT_SUCCESS;

    if (file)
    {
	encode_palette(palrec, *palette);
	sprintf(string, "writing raster palette: file pos:%ld", ftell(file));
	/*muse_log(string);*/
	items_written = fwrite((void *) palrec, sizeof(palrec), (size_t) 1, file);
	if (items_written != 1)
	{
	    printf("Unable to write the palette");
	    return (STAT_FILE_WRITE_ERR);
	}
    }
    if (destruct)
    {
	muse_free(*palette, "palette");
	*palette = (PALETTE *) NULL;
    }
    return (STAT_SUCCESS);
}


#if XVT_CC_PROTO
ERRSTATUS
pal_gen(PALETTE * pal)
#else
ERRSTATUS
pal_gen(pal)
    PALETTE        *pal;
#endif

{
    long            ir, ig, ib, index = 0;

    pal->num_colors = 216;
    pal->magic = MAGIC_PALETTE;

    for (ir = 0; ir < 6; ir++)
    {
	for (ig = 0; ig < 6; ig++)
	{
	    for (ib = 0; ib < 6; ib++)
	    {
		pal->rgb[index].r = (unsigned short)(ir * 10922);
		pal->rgb[index].g = (unsigned short)(ig * 10922);
		pal->rgb[index].b = (unsigned short)(ib * 10922);
		index++;
	    }
	}
    }

    return (STAT_SUCCESS);
}

#if XVT_CC_PROTO
MUSE_API pal_nearest_color(PALETTE * pal, RGB rgb, USHORT * pal_color)
#else
MUSE_API pal_nearest_color(pal, rgb, pal_color)
    PALETTE        *pal;
    RGB             rgb;
    USHORT         *pal_color;
#endif
{
    long            distance, min_distance, i;
    RGB             t;
    switch (pal->type)
    {
      case PAL_UNKNOWN:
	min_distance = LARGE_NUMBER;
	for (i = 0; i < pal->num_colors; i++)
	{
	    distance = abs(rgb.r - pal->rgb[i].r)
		      + abs(rgb.g - pal->rgb[i].g)
		      + abs(rgb.b - pal->rgb[i].b);
	    if (distance < min_distance)
	    {
		min_distance = distance;
		*pal_color = i;
	    }
	}
	break;

      case PAL_BIT_BINS:
	*pal_color = 0;

	/*
	 * the RGB values are right shifted to truncate, then left shifted to
	 * their positions in the color index, then or'd together
	 */
	*pal_color =
		  (rgb.b >> (RGB_BITS - pal->bins.b))
		  | ((rgb.g >> (RGB_BITS - pal->bins.g))
		     << pal->bins.b)
		  | ((rgb.r >> (RGB_BITS - pal->bins.r))
		     << (pal->bins.b + pal->bins.g));
	break;

      case PAL_NUM_BINS:

	/*
	 * the nearest palette color is computed using integer math
	 */
	t.r = min((rgb.r) / (RGB_MAX / pal->bins.r + 0), pal->bins.r - 1);
	t.g = min((rgb.g) / (RGB_MAX / pal->bins.g + 0), pal->bins.r - 1);
	t.b = min((rgb.b) / (RGB_MAX / pal->bins.b + 0), pal->bins.r - 1);
	*pal_color = t.b
		  + t.g * pal->bins.b
		  + t.r * pal->bins.b * pal->bins.g;
	if ((t.r >= pal->bins.r) ||
	    (t.g >= pal->bins.g) ||
	    (t.b >= pal->bins.b) ||
	    (*pal_color > pal->num_colors)
		  )
	    printf("pal_nearest_color %u %u %u %u %u %u \n %X %u",
	    t.r, t.g, t.b, rgb.r, rgb.g, rgb.b, *pal_color, pal->num_colors);

	break;

      default:
	return (-1);
    }
    return STAT_SUCCESS;
}

